# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Project Team-Member Classes."""

from copy import copy

try:
    from typing import Any, Dict, Optional
except ImportError:
    pass  # Python 2.7

from honeybee_ph import _base


class ProjectTeamMember(_base._Base):
    def __init__(
        self,
        _name=None,
        _street=None,
        _city=None,
        _post_code=None,
        _telephone=None,
        _email=None,
        _license_number=None,
    ):
        # type: (Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]) -> None
        super(ProjectTeamMember, self).__init__()
        self.name = _name
        self.street = _street
        self.city = _city
        self.post_code = _post_code
        self.telephone = _telephone
        self.email = _email
        self.license_number = _license_number

    def duplicate(self):
        # type: () -> ProjectTeamMember
        new_obj = ProjectTeamMember(
            self.name,
            self.street,
            self.city,
            self.post_code,
            self.telephone,
            self.email,
            self.license_number,
        )
        new_obj.set_base_attrs_from_source(self)
        return new_obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d["name"] = self.name
        d["street"] = self.street
        d["city"] = self.city
        d["post_code"] = self.post_code
        d["telephone"] = self.telephone
        d["email"] = self.email
        d["license_number"] = self.license_number
        d["identifier"] = self.identifier
        d["user_data"] = self.user_data
        d["display_name"] = self.display_name
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> ProjectTeamMember
        new_obj = cls(
            _input_dict["name"],
            _input_dict["street"],
            _input_dict["city"],
            _input_dict["post_code"],
            _input_dict["telephone"],
            _input_dict["email"],
            _input_dict["license_number"],
        )
        new_obj.identifier = _input_dict.get("identifier", "")
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.display_name = _input_dict.get("display_name", "")
        return new_obj

    def __str__(self):
        return "{}(name={}, street={})".format(self.__class__.__name__, self.name, self.street)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class ProjectTeam(_base._Base):
    def __init__(self):
        # type: () -> None
        super(ProjectTeam, self).__init__()
        self.customer = ProjectTeamMember()
        self.building = ProjectTeamMember()
        self.owner = ProjectTeamMember()
        self.designer = ProjectTeamMember()
        self.project_date = ""
        self.owner_is_client = False
        self.year_constructed = 0
        self.image = None

    def duplicate(self):
        # type: () -> ProjectTeam
        new_obj = ProjectTeam()
        new_obj.set_base_attrs_from_source(self)
        new_obj.customer = self.customer.duplicate()
        new_obj.building = self.building.duplicate()
        new_obj.owner = self.owner.duplicate()
        new_obj.designer = self.designer.duplicate()
        return new_obj

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d["identifier"] = self.identifier
        d["user_data"] = self.user_data
        d["display_name"] = self.display_name
        d["customer"] = self.customer.to_dict()
        d["building"] = self.building.to_dict()
        d["owner"] = self.owner.to_dict()
        d["designer"] = self.designer.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> ProjectTeam
        new_obj = cls()
        new_obj.identifier = _input_dict.get("identifier", "")
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.display_name = _input_dict.get("display_name", "")
        new_obj.customer = ProjectTeamMember.from_dict(_input_dict.get("customer", ProjectTeamMember()))
        new_obj.building = ProjectTeamMember.from_dict(_input_dict.get("building", ProjectTeamMember()))
        new_obj.owner = ProjectTeamMember.from_dict(_input_dict.get("owner", ProjectTeamMember()))
        new_obj.designer = ProjectTeamMember.from_dict(_input_dict.get("designer", ProjectTeamMember()))
        return new_obj



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Passive House properties for honeybee_energy.construction.opaque.OpaqueConstruction Objects"""


class OpaqueConstructionPhProperties(object):
    def __init__(self):
        self.id_num = 0

    def __repr__(self):
        return "{}(id_num={!r})".format(self.__class__.__name__, self.id_num)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Heating Devices."""

import sys
from copy import copy

try:
    from typing import Any, Optional
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee_phhvac import _base, fuels
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


class UnknownPhHeatingTypeError(Exception):
    def __init__(self, _heater_types, _received_type):
        # type: (list[str], str) -> None
        self.msg = 'Error: Unknown HBPH-Heating-SubSystem type? Got: "{}" but only types: {} are allowed?'.format(
            _received_type, _heater_types
        )
        super(UnknownPhHeatingTypeError, self).__init__(self.msg)


class PhHeatingSystem(_base._PhHVACBase):
    """Base class for all PH-Heating Systems (elec, boiler, etc...)"""

    def __init__(self):
        super(PhHeatingSystem, self).__init__()
        self.heating_type = self.__class__.__name__
        self.percent_coverage = 1.0

    def to_dict(self):
        # type: () -> dict
        d = super(PhHeatingSystem, self).to_dict()
        d["heating_type"] = self.heating_type
        d["percent_coverage"] = self.percent_coverage
        return d

    def base_attrs_from_dict(self, _input_dict):
        # type: (PhHeatingSystem, dict) -> PhHeatingSystem
        self.identifier = _input_dict["identifier"]
        self.display_name = _input_dict["display_name"]
        self.user_data = _input_dict["user_data"]
        self.heating_type = _input_dict["heating_type"]
        self.percent_coverage = _input_dict["percent_coverage"]
        return self

    def check_dict_type(self, _input_dict):
        # type: (dict) -> None
        """Check that the input dict type is correct for the Heating System being constructed."""
        heating_type = _input_dict["heating_type"]
        msg = "Error creating Heating System from dict. Expected '{}' but got '{}'".format(
            self.__class__.__name__, heating_type
        )
        assert heating_type == str(self.__class__.__name__), msg
        return None

    @classmethod
    def from_dict(cls, _input_dict):
        raise NotImplementedError("Error: from_dict() called on BaseClass.")

    def __lt__(self, other):
        # type: (PhHeatingSystem) -> bool
        return self.identifier < other.identifier

    def move(self, moving_vec3D):
        """Move the System's elements along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the ray.
        """
        pass

    def rotate(self, axis_vec3D, angle_degree, origin_pt3D):
        """Rotate the System's elements by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in degrees.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        pass

    def rotate_xy(self, angle_degree, origin_pt3D):
        """Rotate the System's elements counterclockwise in the XY plane by a certain angle.

        Args:
            angle: An angle in degrees.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        pass

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the System's elements across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        pass

    def scale(self, scale_factor, origin=None):
        """Scale the System's elements by a factor from an origin point.

        Args:
            factor: A number representing how much the line segment should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        pass

    def duplicate(self):
        # type: () -> PhHeatingSystem
        raise NotImplementedError("Error: duplicate() called on BaseClass.")

    def __copy__(self):
        raise NotImplementedError("Error: __copy__() called on BaseClass.")


# -----------------------------------------------------------------------------
# Heating Types


class PhHeatingDirectElectric(PhHeatingSystem):
    """Heating via direct-electric (resistance heating)."""

    def __init__(self):
        super(PhHeatingDirectElectric, self).__init__()

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = super(PhHeatingDirectElectric, self).to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhHeatingDirectElectric
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatingDirectElectric
        obj = PhHeatingDirectElectric()
        obj.identifier = self.identifier
        obj.display_name = self.display_name
        obj.user_data = copy(self.user_data)
        return obj


class PhHeatingFossilBoiler(PhHeatingSystem):
    """Heating via boiler using fossil-fuel (gas, oil)"""

    def __init__(self):
        super(PhHeatingFossilBoiler, self).__init__()
        self.fuel = fuels.NATURAL_GAS
        self.condensing = True
        self.in_conditioned_space = True
        self.effic_at_30_percent_load = 0.98
        self.effic_at_nominal_load = 0.94
        self.avg_rtrn_temp_at_30_percent_load = 30
        self.avg_temp_at_70C_55C = 41
        self.avg_temp_at_55C_45C = 35
        self.avg_temp_at_32C_28C = 24

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = super(PhHeatingFossilBoiler, self).to_dict()

        d["fuel"] = self.fuel
        d["condensing"] = self.condensing
        d["in_conditioned_space"] = self.in_conditioned_space
        d["effic_at_30_percent_load"] = self.effic_at_30_percent_load
        d["effic_at_nominal_load"] = self.effic_at_nominal_load
        d["avg_rtrn_temp_at_30_percent_load"] = self.avg_rtrn_temp_at_30_percent_load
        d["avg_temp_at_70C_55C"] = self.avg_temp_at_70C_55C
        d["avg_temp_at_55C_45C"] = self.avg_temp_at_55C_45C
        d["avg_temp_at_32C_28C"] = self.avg_temp_at_32C_28C

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhHeatingFossilBoiler
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)

        new_obj.fuel = _input_dict["fuel"]
        new_obj.condensing = _input_dict["condensing"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.effic_at_30_percent_load = _input_dict["effic_at_30_percent_load"]
        new_obj.effic_at_nominal_load = _input_dict["effic_at_nominal_load"]
        new_obj.avg_rtrn_temp_at_30_percent_load = _input_dict["avg_rtrn_temp_at_30_percent_load"]
        new_obj.avg_temp_at_70C_55C = _input_dict["avg_temp_at_70C_55C"]
        new_obj.avg_temp_at_55C_45C = _input_dict["avg_temp_at_55C_45C"]
        new_obj.avg_temp_at_32C_28C = _input_dict["avg_temp_at_32C_28C"]

        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatingFossilBoiler
        obj = PhHeatingFossilBoiler()
        obj.identifier = self.identifier
        obj.display_name = self.display_name
        obj.user_data = copy(self.user_data)
        obj.fuel = self.fuel
        obj.condensing = self.condensing
        obj.in_conditioned_space = self.in_conditioned_space
        obj.effic_at_30_percent_load = self.effic_at_30_percent_load
        obj.effic_at_nominal_load = self.effic_at_nominal_load
        obj.avg_rtrn_temp_at_30_percent_load = self.avg_rtrn_temp_at_30_percent_load
        obj.avg_temp_at_70C_55C = self.avg_temp_at_70C_55C
        obj.avg_temp_at_55C_45C = self.avg_temp_at_55C_45C
        obj.avg_temp_at_32C_28C = self.avg_temp_at_32C_28C
        return obj


class PhHeatingWoodBoiler(PhHeatingSystem):
    """Heating via boiler using wood (log, pellet)."""

    def __init__(self):
        super(PhHeatingWoodBoiler, self).__init__()
        self.fuel = fuels.WOOD_LOG
        self.in_conditioned_space = True
        self.effic_in_basic_cycle = 0.6
        self.effic_in_const_operation = 0.7
        self.avg_frac_heat_output = 0.4
        self.temp_diff_on_off = 30
        self.rated_capacity = 15  # kW
        self.demand_basic_cycle = 1  # kWh
        self.power_stationary_run = 1  # W
        self.power_standard_run = None  # type: Optional[float]
        self.no_transport_pellets = None  # type: Optional[bool]
        self.only_control = None  # type: Optional[bool]
        self.area_mech_room = None  # type: Optional[float]

    @property
    def useful_heat_output(self):
        return 0.9 * self.rated_capacity  # kWH

    @property
    def avg_power_output(self):
        return 0.5 * self.rated_capacity  # kW

    def to_dict(self):
        # type: () -> dict
        d = super(PhHeatingWoodBoiler, self).to_dict()

        d["fuel"] = self.fuel
        d["in_conditioned_space"] = self.in_conditioned_space
        d["effic_in_basic_cycle"] = self.effic_in_basic_cycle
        d["effic_in_const_operation"] = self.effic_in_const_operation
        d["avg_frac_heat_output"] = self.avg_frac_heat_output
        d["temp_diff_on_off"] = self.temp_diff_on_off
        d["rated_capacity"] = self.rated_capacity
        d["demand_basic_cycle"] = self.demand_basic_cycle
        d["power_stationary_run"] = self.power_stationary_run
        d["power_standard_run"] = self.power_standard_run
        d["no_transport_pellets"] = self.no_transport_pellets
        d["only_control"] = self.only_control
        d["area_mech_room"] = self.area_mech_room

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHeatingWoodBoiler

        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)

        new_obj.fuel = _input_dict["fuel"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.effic_in_basic_cycle = _input_dict["effic_in_basic_cycle"]
        new_obj.effic_in_const_operation = _input_dict["effic_in_const_operation"]
        new_obj.avg_frac_heat_output = _input_dict["avg_frac_heat_output"]
        new_obj.temp_diff_on_off = _input_dict["temp_diff_on_off"]
        new_obj.rated_capacity = _input_dict["rated_capacity"]
        new_obj.demand_basic_cycle = _input_dict["demand_basic_cycle"]
        new_obj.power_stationary_run = _input_dict["power_stationary_run"]
        new_obj.no_transport_pellets = _input_dict["no_transport_pellets"]
        new_obj.only_control = _input_dict["only_control"]
        new_obj.area_mech_room = _input_dict["area_mech_room"]

        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatingWoodBoiler
        obj = PhHeatingWoodBoiler()
        obj.identifier = self.identifier
        obj.display_name = self.display_name
        obj.user_data = copy(self.user_data)
        obj.fuel = self.fuel
        obj.in_conditioned_space = self.in_conditioned_space
        obj.effic_in_basic_cycle = self.effic_in_basic_cycle
        obj.effic_in_const_operation = self.effic_in_const_operation
        obj.avg_frac_heat_output = self.avg_frac_heat_output
        obj.temp_diff_on_off = self.temp_diff_on_off
        obj.rated_capacity = self.rated_capacity
        obj.demand_basic_cycle = self.demand_basic_cycle
        obj.power_stationary_run = self.power_stationary_run
        obj.power_standard_run = self.power_standard_run
        obj.no_transport_pellets = self.no_transport_pellets
        obj.only_control = self.only_control
        obj.area_mech_room = self.area_mech_room
        return obj


class PhHeatingDistrict(PhHeatingSystem):
    """Heating via district-heat."""

    def __init__(self):
        super(PhHeatingDistrict, self).__init__()
        self.fuel = fuels.GAS_CGS_70_PHC
        self.util_factor_of_heat_transfer_station = 0.5

    def to_dict(self):
        # type: () -> dict
        d = super(PhHeatingDistrict, self).to_dict()
        d["fuel"] = self.fuel
        d["util_factor_of_heat_transfer_station"] = self.util_factor_of_heat_transfer_station
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHeatingDistrict
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)

        new_obj.fuel = _input_dict["fuel"]
        new_obj.util_factor_of_heat_transfer_station = _input_dict["util_factor_of_heat_transfer_station"]
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatingDistrict
        obj = PhHeatingDistrict()
        obj.identifier = self.identifier
        obj.display_name = self.display_name
        obj.user_data = copy(self.user_data)
        obj.fuel = self.fuel
        obj.util_factor_of_heat_transfer_station = self.util_factor_of_heat_transfer_station
        return obj


# -----------------------------------------------------------------------------


class PhHeatingSystemBuilder(object):
    """Constructor class for PH-HeatingSystems"""

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhHeatingSystem
        """Find the right appliance constructor class from the module based on the 'type' name."""
        valid_class_types = [nm for nm in dir(sys.modules[__name__]) if nm.startswith("Ph")]

        heating_type = _input_dict["heating_type"]
        if heating_type not in valid_class_types:
            raise UnknownPhHeatingTypeError(valid_class_types, heating_type)
        heating_class = getattr(sys.modules[__name__], heating_type)
        new_equipment = heating_class.from_dict(_input_dict)
        return new_equipment

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-PH-HVAC Aperture Properties."""

try:
    from typing import Any, Dict, Optional
except ImportError:
    pass  # Python 2.7

try:
    from honeybee.properties import ApertureProperties
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))


class AperturePhHvacProperties(object):
    def __init__(self, _host):
        # type: (Optional[ApertureProperties]) -> None
        self._host = _host

    @property
    def host(self):
        # type: () -> Optional[ApertureProperties]
        return self._host

    def __copy__(self, new_host=None):
        # type: (Optional[ApertureProperties]) -> AperturePhHvacProperties
        _host = new_host or self._host
        new_properties_obj = AperturePhHvacProperties(_host)
        return new_properties_obj

    def duplicate(self, new_host=None):
        # type: (Any) -> AperturePhHvacProperties
        return self.__copy__(new_host=new_host)

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        if abridged == False:
            d["type"] = "AperturePhHvacPropertiesAbridged"
        else:
            d["type"] = "AperturePhHvacProperties"
        return {"ph_hvac": d}

    @classmethod
    def from_dict(cls, _dict, host):
        # type: (dict[str, Any], Any) -> AperturePhHvacProperties
        assert _dict["type"] == "AperturePhHvacProperties", "Expected AperturePhHvacProperties. Got {}.".format(
            _dict["type"]
        )
        new_prop = cls(host)
        return new_prop

    def apply_properties_from_dict(self, data):
        # type: (Dict[str, Any]) -> None
        return None



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Passive House properties for honeybee_energy.material.opaque.EnergyMaterial Objects"""


try:
    from typing import Any, Iterable, List, NoReturn, Optional
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee_energy.material import opaque
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_energy:\n\t{}".format(e))

try:
    from honeybee_ph_utils.color import PhColor
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph_utils:\n\t{}".format(e))


class CellPositionError(Exception):
    """An error for when a cell position is out of range."""

    def __init__(self, msg):
        super(CellPositionError, self).__init__(msg)


class PhDivisionCell(object):
    """A single cell in a PhLayerDivisionGrid."""

    def __init__(self, _row, _column, _hbe_material):
        # type: (int, int, opaque.EnergyMaterial) -> None
        self.row = _row
        self.column = _column
        self.material = _hbe_material

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = {}
        d["row"] = self.row
        d["column"] = self.column
        d["material"] = self.material.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhDivisionCell
        new_cell = cls(
            _input_dict["row"], _input_dict["column"], opaque.EnergyMaterial.from_dict(_input_dict["material"])
        )
        return new_cell

    def duplicate(self):
        # type: () -> PhDivisionCell
        """Duplicate the cell."""
        return PhDivisionCell(self.row, self.column, self.material.duplicate())

    def __copy__(self):
        # type: () -> PhDivisionCell
        return self.duplicate()

    def __str__(self):
        return "{}(row={}, column={}, material={})".format(
            self.__class__.__name__,
            self.row,
            self.column,
            self.material.display_name,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class PhDivisionGrid(object):
    """A grid of PhDivisionCell to support 'mixed' materials.

    The Cell grid is ordered from top-left to bottom-right:

    |    | C0  | C1  | C2  | ...
    |:---|:---:|:---:|:---:|:---:
    | R0 | 0,0 | 1,0 | 2,0 | ...
    | R1 | 0,1 | 1,1 | 2,1 | ...
    | R2 | 0,2 | 1,2 | 2,2 | ...
    """

    def __init__(self):
        # type: () -> None
        self._row_heights = []  # type: List[float]
        self._column_widths = []  # type: List[float]
        self._cells = []  # type: List[PhDivisionCell]

    @property
    def column_widths(self):
        # type: () -> List[float]
        """Return the list of column widths."""
        return self._column_widths

    @property
    def column_count(self):
        # type: () -> int
        """Return the number of columns in the grid."""
        return len(self._column_widths)

    @property
    def row_heights(self):
        # type: () -> List[float]
        """Return the list of row heights."""
        return self._row_heights

    @property
    def row_count(self):
        # type: () -> int
        """Return the number of rows in the grid."""
        return len(self._row_heights)

    @property
    def cell_count(self):
        # type: () -> int
        """Return the total number of cells in the grid."""
        return self.row_count * self.column_count

    @property
    def cells(self):
        # type: () -> List[PhDivisionCell]
        """Return the list of all the cells in the grid."""
        return self._cells

    def set_column_widths(self, _column_widths):
        # type: (Iterable[float]) -> None
        """Set the column widths of the grid."""
        self._column_widths = []
        for width in _column_widths:
            self.add_new_column(width)

    def add_new_column(self, _column_width):
        # type: (float) -> None
        """Add a new COLUMN to the grid with the given width."""
        if _column_width is None:
            return

        if _column_width > 0:
            self._column_widths.append(float(_column_width))

    def set_row_heights(self, _row_heights):
        # type: (Iterable[float]) -> None
        """Set the row heights of the grid."""
        self._row_heights = []
        for height in _row_heights:
            self.add_new_row(height)

    def add_new_row(self, _row_height):
        # type: (float) -> None
        """Add a new ROW to the grid with the given height."""
        if _row_height is None:
            return

        if _row_height > 0:
            self._row_heights.append(float(_row_height))

    def get_cell(self, _column, _row):
        # type: (int, int) -> Optional[PhDivisionCell]
        """Get the PhDivisionCell at the given column and row position."""
        for cell in self._cells:
            if cell.column == _column and cell.row == _row:
                return cell
        return None

    def set_cell_material(self, _column_num, _row_num, _hbe_material):
        # type: (int, int, opaque.EnergyMaterial) -> None
        """Set the EnergyMaterial for a specific cell in the grid by its column/row position.

        Cells are indexed by their column and row position stating from top-left:

        |   | C0   | C1  | C2  | ...
        |---|------|-----|-----|----
        |R0 | 0,0  | 1,0 | 2,0 | ...
        |R1 | 0,1  | 1,1 | 2,1 | ...
        |R2 | 0,2  | 1,2 | 2,2 | ...
        """
        if _column_num >= self.column_count:
            msg = (
                "Error setting Material '{}' to column '{}'. The specified column is out of range. "
                "Please setup all the column-widths before assigning any materials. And "
                "remember that the columns start counting from '0', not '1'.".format(
                    _hbe_material.display_name, _column_num
                )
            )
            raise CellPositionError(msg)

        if _row_num >= self.row_count:
            msg = (
                "Error setting Material '{}' to row '{}'. The specified row is out of range. "
                "Please setup all the row-heights before assigning any materials. And "
                "remember that the rows start counting from '0', not '1'.".format(_hbe_material.display_name, _row_num)
            )
            raise CellPositionError(msg)

        # -- See if the cell already exists, if so reset its material
        # -- if not, create a new cell.
        cell = self.get_cell(_column_num, _row_num)
        if cell:
            cell.material = _hbe_material
        else:
            cell = PhDivisionCell(_row=_row_num, _column=_column_num, _hbe_material=_hbe_material)

        self._cells.append(cell)

    def get_cell_material(self, _column_num, _row_num):
        # type: (int, int) -> Optional[opaque.EnergyMaterial]
        """Get the PhxMaterial for a specific cell in the grid by its column/row position."""
        for cell in self._cells:
            if cell.row == _row_num and cell.column == _column_num:
                return cell.material
        return None

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = {}
        d["column_widths"] = self.column_widths
        d["row_heights"] = self.row_heights
        d["cells"] = []
        for cell in self._cells:
            d["cells"].append(cell.to_dict())
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhDivisionGrid
        new_grid = cls()
        new_grid.set_column_widths(_input_dict.get("column_widths", []))
        new_grid.set_row_heights(_input_dict.get("row_heights", []))
        for cell_dict in _input_dict.get("cells", []):
            new_grid.set_cell_material(
                cell_dict["column"],
                cell_dict["row"],
                opaque.EnergyMaterial.from_dict(cell_dict["material"]),
            )
        return new_grid

    def duplicate(self):
        # type: () -> PhDivisionGrid
        """Duplicate the grid."""
        new_grid = PhDivisionGrid()
        new_grid.set_column_widths(self.column_widths)
        new_grid.set_row_heights(self.row_heights)
        for cell in self._cells:
            new_grid.set_cell_material(cell.column, cell.row, cell.material.duplicate())
        return new_grid

    def __copy__(self):
        # type: () -> PhDivisionGrid
        return self.duplicate()

    def __str__(self):
        return "{}(columns={}, rows={})".format(
            self.__class__.__name__,
            self.column_count,
            self.row_count,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class EnergyMaterialPhProperties(object):
    """Passive House properties for EnergyMaterial objects."""

    def __init__(self, _host=None):
        # type: (Optional[opaque.EnergyMaterial]) -> None
        self._host = _host
        self.id_num = 0
        self._ph_color = None  # type: Optional[PhColor]
        self.user_data = {}
        self.base_material = None  # type: Optional[opaque.EnergyMaterial]
        self.divisions = PhDivisionGrid()

    # -------------------------------------------------------------------------
    # ------------------ Deprecated [April 4, 2024] ---------------------------

    @property
    def percentage_of_assembly(self):
        # type: () -> NoReturn
        raise DeprecationWarning(
            "The 'percentage_of_assembly' property is deprecated. Please use the 'divisions' for mixed materials."
        )

    @percentage_of_assembly.setter
    def percentage_of_assembly(self, _percentage):
        # type: (Any) -> NoReturn
        raise DeprecationWarning(
            "The 'percentage_of_assembly' property is deprecated. Please use the 'divisions' for mixed materials."
        )

    @property
    def base_materials(self):
        # type: () -> NoReturn
        raise DeprecationWarning(
            "The 'base_materials' property is deprecated. Please use the 'divisions' for mixed materials."
        )

    @base_materials.setter
    def base_materials(self, _materials):
        # type: (Any) -> NoReturn
        raise DeprecationWarning(
            "The 'base_materials' property is deprecated. Please use the 'divisions' for mixed materials."
        )

    def add_base_material(self, _hb_material):
        # type: (Any) -> NoReturn
        raise DeprecationWarning(
            "The 'add_base_material' function is deprecated. Please use the 'divisions' for mixed materials."
        )

    def clear_base_materials(self):
        # type: (Any) -> NoReturn
        raise DeprecationWarning(
            "The 'clear_base_materials' function is deprecated. Please use the 'divisions' for mixed materials."
        )

    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------

    @property
    def ph_color(self):
        # type: () -> Optional[PhColor]
        return self._ph_color

    @ph_color.setter
    def ph_color(self, _input_color):
        # type: (Optional[PhColor]) -> None
        self._ph_color = _input_color

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        d["id_num"] = self.id_num
        d["divisions"] = self.divisions.to_dict()
        d["user_data"] = self.user_data

        if self.ph_color:
            d["ph_color"] = self.ph_color.to_dict()

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, _host):
        # type: (dict, Optional[opaque.EnergyMaterial]) -> EnergyMaterialPhProperties
        new_prop = cls(_host)
        new_prop.id_num = _input_dict["id_num"]
        new_prop._ph_color = PhColor.from_dict(_input_dict.get("ph_color", None))
        new_prop.user_data = _input_dict.get("user_data", {})
        new_prop.divisions = PhDivisionGrid.from_dict(_input_dict.get("divisions", {}))
        return new_prop

    def apply_properties_from_dict(self, abridged_data):
        # type: (Any) -> None
        return None

    def __copy__(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterial]) -> EnergyMaterialPhProperties
        _host = new_host or self._host
        new_obj = EnergyMaterialPhProperties(_host)
        new_obj.id_num = self.id_num
        new_obj.divisions = self.divisions.duplicate()
        new_obj.user_data = self.user_data.copy()
        if self.ph_color:
            new_obj.ph_color = self.ph_color.duplicate()
        return new_obj

    def duplicate(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterial]) -> EnergyMaterialPhProperties
        return self.__copy__(new_host)

    def __str__(self):
        return "{}(id_num={!r}, ph_color={!r})".format(
            self.__class__.__name__,
            self.id_num,
            self.ph_color,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class EnergyMaterialNoMassPhProperties(object):
    def __init__(self, _host=None):
        # type: (Optional[opaque.EnergyMaterialNoMass]) -> None
        self.host = _host
        self.id_num = 0
        self._ph_color = None  # type: Optional[PhColor]
        self.user_data = {}

    @property
    def ph_color(self):
        # type: () -> Optional[PhColor]
        return self._ph_color

    @ph_color.setter
    def ph_color(self, _input_color):
        # type: (Optional[PhColor]) -> None
        self._ph_color = _input_color

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        d["id_num"] = self.id_num
        d["user_data"] = self.user_data

        if self.ph_color:
            d["ph_color"] = self.ph_color.to_dict()

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, _host):
        # type: (dict, Optional[opaque.EnergyMaterialNoMass]) -> EnergyMaterialNoMassPhProperties
        new_prop = cls(_host)
        new_prop.id_num = _input_dict["id_num"]
        new_prop._ph_color = PhColor.from_dict(_input_dict.get("ph_color", None))
        new_prop.user_data = _input_dict.get("user_data", {})
        return new_prop

    def apply_properties_from_dict(self, abridged_data):
        # type: (dict) -> None
        return

    def __copy__(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterialNoMass]) -> EnergyMaterialNoMassPhProperties
        _host = new_host or self.host
        new_obj = EnergyMaterialNoMassPhProperties(_host)
        new_obj.id_num = self.id_num
        new_obj.user_data = self.user_data.copy()
        if self.ph_color:
            new_obj.ph_color = self.ph_color.duplicate()
        return new_obj

    def duplicate(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterialNoMass]) -> EnergyMaterialNoMassPhProperties
        return self.__copy__(new_host)

    def __str__(self):
        return "{}(id_num={!r}, ph_color={!r})".format(self.__class__.__name__, self.id_num, self.ph_color)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class EnergyMaterialVegetationPhProperties(object):
    def __init__(self, _host=None):
        # type: (Optional[opaque.EnergyMaterialVegetation]) -> None
        self.host = _host
        self.id_num = 0
        self._ph_color = None  # type: Optional[PhColor]
        self.user_data = {}

    @property
    def ph_color(self):
        # type: () -> Optional[PhColor]
        return self._ph_color

    @ph_color.setter
    def ph_color(self, _input_color):
        # type: (Optional[PhColor]) -> None
        self._ph_color = _input_color

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        d["id_num"] = self.id_num
        d["user_data"] = self.user_data

        if self.ph_color:
            d["ph_color"] = self.ph_color.to_dict()

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, _host):
        # type: (dict[str, Any], Optional[opaque.EnergyMaterialVegetation]) -> EnergyMaterialVegetationPhProperties
        new_prop = cls(_host)
        new_prop.id_num = _input_dict["id_num"]
        new_prop._ph_color = PhColor.from_dict(_input_dict.get("ph_color", None))
        new_prop.user_data = _input_dict.get("user_data", {})
        return new_prop

    def apply_properties_from_dict(self, abridged_data):
        return

    def __copy__(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterialVegetation]) -> EnergyMaterialVegetationPhProperties
        _host = new_host or self.host
        new_obj = EnergyMaterialVegetationPhProperties(_host)
        new_obj.id_num = self.id_num
        new_obj.user_data = self.user_data.copy()
        if self.ph_color:
            new_obj.ph_color = self.ph_color.duplicate()
        return new_obj

    def duplicate(self, new_host=None):
        # type: (Optional[opaque.EnergyMaterialVegetation]) -> EnergyMaterialVegetationPhProperties
        return self.__copy__(new_host)

    def __str__(self):
        return "{}(id_num={!r}, ph_color={!r})".format(self.__class__.__name__, self.id_num, self.ph_color)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Heat-Pump Devices."""

import sys

try:
    from typing import Any, Dict, List, Sequence
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee_phhvac import _base
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))

# -----------------------------------------------------------------------------
# Heat Pump Base


class UnknownPhHeatPumpTypeError(Exception):
    def __init__(self, _heater_types, _received_type):
        # type: (List[str], str) -> None
        self.msg = 'Error: Unknown HBPH-Heat-pump system type? Got: "{}" but only types: {} are allowed?'.format(
            _received_type, _heater_types
        )
        super(UnknownPhHeatPumpTypeError, self).__init__(self.msg)


class PhHeatPumpSystem(_base._PhHVACBase):
    """Base class for all HBPH-Cooling-Systems."""

    def __init__(self):
        super(PhHeatPumpSystem, self).__init__()
        self.heat_pump_class_name = self.__class__.__name__
        self.percent_coverage = 1.0
        self.cooling_params = PhHeatPumpCoolingParams()

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpSystem, self).to_dict()
        d["heat_pump_class_name"] = self.heat_pump_class_name
        d["percent_coverage"] = self.percent_coverage
        return d

    def base_attrs_from_dict(self, _input_dict):
        # type: (PhHeatPumpSystem, Dict[str, Any]) -> PhHeatPumpSystem
        self.identifier = _input_dict["identifier"]
        self.display_name = _input_dict["display_name"]
        self.user_data = _input_dict.get("user_data", {})
        self.heat_pump_class_name = _input_dict["heat_pump_class_name"]
        self.percent_coverage = _input_dict["percent_coverage"]
        return self

    def check_dict_type(self, _input_dict):
        # type: (Dict[str, Any]) -> None
        """Check that the input dict type is correct for the Heat Pump System being constructed."""
        heat_pump_class_name = _input_dict["heat_pump_class_name"]
        msg = "Error creating Heat Pump System from dict. Expected '{}' but got '{}'".format(
            self.__class__.__name__, heat_pump_class_name
        )
        assert heat_pump_class_name == str(self.__class__.__name__), msg
        return None

    @classmethod
    def from_dict(cls, input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpSystem
        raise NotImplementedError("Error: from_dict() called from BaseClass")

    def __lt__(self, other):
        # type: (PhHeatPumpSystem) -> bool
        return self.identifier < other.identifier

    def move(self, moving_vec3D):
        """Move the System's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        pass

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        """Rotate the System's elements by a certain angle around an axis_vec3D and origin_pt3D.

        Right hand rule applies:
        If axis_vec3D has a positive orientation, rotation will be clockwise.
        If axis_vec3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis_vec3D representing the axis_vec3D of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def rotate_xy(self, angle_degrees, origin_pt3D):
        """Rotate the System's elements counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        """
        pass

    def reflect(self, normal_vec3D, origin_pt3D):
        """Reflected the System's elements across a plane with the input normal vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        """
        pass

    def scale(self, scale_factor, origin_pt3D=None):
        """Scale the System's elements by a factor from an origin_pt3D point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        """
        pass

    def __copy__(self):
        # type: () -> PhHeatPumpSystem
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpSystem
        """Duplicate the System."""
        raise NotImplementedError("Error: duplicate() called from BaseClass")


# -----------------------------------------------------------------------------
# Heat-Pump Cooling Parameters


class PhHeatPumpCoolingParams_Base(_base._PhHVACBase):
    """Base class for all HBPH-Cooling-Parameters."""

    def __init__(self):
        super(PhHeatPumpCoolingParams_Base, self).__init__()
        self.used = False

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpCoolingParams_Base, self).to_dict()
        d["used"] = self.used
        return d

    def base_attrs_from_dict(self, _input_dict):
        # type: (PhHeatPumpCoolingParams_Base, Dict[str, Any]) -> PhHeatPumpCoolingParams_Base
        self.identifier = _input_dict["identifier"]
        self.display_name = _input_dict["display_name"]
        self.user_data = _input_dict.get("user_data", {})
        self.used = _input_dict["used"]
        return self

    @classmethod
    def from_dict(cls, input_dict):
        raise NotImplementedError("Error: from_dict() called from PhHeatPumpCoolingParams_Base?")

    def __copy__(self):
        raise NotImplementedError("Error: __copy__() called from PhHeatPumpCoolingParams_Base?")

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams_Base
        raise NotImplementedError("Error: duplicate() called from PhHeatPumpCoolingParams_Base?")

    def __str__(self):
        # type: () -> str
        return "{}(used={})".format(self.__class__.__name__, self.used)


class PhHeatPumpCoolingParams_Ventilation(PhHeatPumpCoolingParams_Base):
    """Cooling via the Fresh-Air Ventilation System (ERV)."""

    def __init__(self):
        super(PhHeatPumpCoolingParams_Ventilation, self).__init__()
        self.single_speed = False
        self.min_coil_temp = 12.0
        self.capacity = 10.0
        self.annual_COP = 4.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpCoolingParams_Ventilation, self).to_dict()
        d["single_speed"] = self.single_speed
        d["min_coil_temp"] = self.min_coil_temp
        d["capacity"] = self.capacity
        d["annual_COP"] = self.annual_COP
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpCoolingParams_Ventilation
        new_obj = cls()
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.single_speed = _input_dict["single_speed"]
        new_obj.min_coil_temp = _input_dict["min_coil_temp"]
        new_obj.capacity = _input_dict["capacity"]
        new_obj.annual_COP = _input_dict["annual_COP"]
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams_Ventilation
        new_obj = PhHeatPumpCoolingParams_Ventilation()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.used = self.used
        new_obj.single_speed = self.single_speed
        new_obj.min_coil_temp = self.min_coil_temp
        new_obj.capacity = self.capacity
        new_obj.annual_COP = self.annual_COP
        return new_obj


class PhHeatPumpCoolingParams_Recirculation(PhHeatPumpCoolingParams_Base):
    """Cooling via a 'recirculation' system (typical AC)."""

    def __init__(self):
        super(PhHeatPumpCoolingParams_Recirculation, self).__init__()
        self.single_speed = False
        self.min_coil_temp = 12.0
        self.flow_rate_m3_hr = 100.0
        self.flow_rate_variable = True
        self.capacity = 10.0
        self.annual_COP = 4.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpCoolingParams_Recirculation, self).to_dict()
        d["single_speed"] = self.single_speed
        d["min_coil_temp"] = self.min_coil_temp
        d["flow_rate_m3_hr"] = self.flow_rate_m3_hr
        d["flow_rate_variable"] = self.flow_rate_variable
        d["capacity"] = self.capacity
        d["annual_COP"] = self.annual_COP
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpCoolingParams_Recirculation
        new_obj = cls()
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.single_speed = _input_dict["single_speed"]
        new_obj.min_coil_temp = _input_dict["min_coil_temp"]
        new_obj.flow_rate_m3_hr = _input_dict["flow_rate_m3_hr"]
        new_obj.flow_rate_variable = _input_dict["flow_rate_variable"]
        new_obj.capacity = _input_dict["capacity"]
        new_obj.annual_COP = _input_dict["annual_COP"]
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams_Recirculation
        new_obj = PhHeatPumpCoolingParams_Recirculation()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.used = self.used
        new_obj.single_speed = self.single_speed
        new_obj.min_coil_temp = self.min_coil_temp
        new_obj.flow_rate_m3_hr = self.flow_rate_m3_hr
        new_obj.flow_rate_variable = self.flow_rate_variable
        new_obj.capacity = self.capacity
        new_obj.annual_COP = self.annual_COP
        return new_obj


class PhHeatPumpCoolingParams_Dehumidification(PhHeatPumpCoolingParams_Base):
    """Cooling via dedicated dehumidification system."""

    def __init__(self):
        super(PhHeatPumpCoolingParams_Dehumidification, self).__init__()
        self.useful_heat_loss = False
        self.annual_COP = 4.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpCoolingParams_Dehumidification, self).to_dict()
        d["useful_heat_loss"] = self.useful_heat_loss
        d["annual_COP"] = self.annual_COP
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpCoolingParams_Dehumidification
        new_obj = cls()
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.useful_heat_loss = _input_dict["useful_heat_loss"]
        new_obj.annual_COP = _input_dict["annual_COP"]
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams_Dehumidification
        new_obj = PhHeatPumpCoolingParams_Dehumidification()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.used = self.used
        new_obj.useful_heat_loss = self.useful_heat_loss
        new_obj.annual_COP = self.annual_COP
        return new_obj


class PhHeatPumpCoolingParams_Panel(PhHeatPumpCoolingParams_Base):
    """Cooling via radiant panels."""

    def __init__(self):
        super(PhHeatPumpCoolingParams_Panel, self).__init__()
        self.annual_COP = 4.0

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpCoolingParams_Panel, self).to_dict()
        d["annual_COP"] = self.annual_COP
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpCoolingParams_Panel
        new_obj = cls()
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.annual_COP = _input_dict["annual_COP"]
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams_Panel
        new_obj = PhHeatPumpCoolingParams_Panel()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.used = self.used
        new_obj.annual_COP = self.annual_COP
        return new_obj


class PhHeatPumpCoolingParams:
    """A Collection of Cooling Parameters for various types of systems."""

    def __init__(self):
        self.percent_coverage = 1.0
        self.ventilation = PhHeatPumpCoolingParams_Ventilation()
        self.recirculation = PhHeatPumpCoolingParams_Recirculation()
        self.dehumidification = PhHeatPumpCoolingParams_Dehumidification()
        self.panel = PhHeatPumpCoolingParams_Panel()

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}
        d["percent_coverage"] = self.percent_coverage
        d["ventilation"] = self.ventilation.to_dict()
        d["recirculation"] = self.recirculation.to_dict()
        d["dehumidification"] = self.dehumidification.to_dict()
        d["panel"] = self.panel.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpCoolingParams
        new_obj = cls()
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.ventilation = PhHeatPumpCoolingParams_Ventilation.from_dict(_input_dict["ventilation"])
        new_obj.recirculation = PhHeatPumpCoolingParams_Recirculation.from_dict(_input_dict["recirculation"])
        new_obj.dehumidification = PhHeatPumpCoolingParams_Dehumidification.from_dict(_input_dict["dehumidification"])
        new_obj.panel = PhHeatPumpCoolingParams_Panel.from_dict(_input_dict["panel"])
        return new_obj

    def __str__(self):
        # type: () -> str
        return "{}(percent_coverage={}, ventilation={}, recirculation={}, dehumidification={}, panel={})".format(
            self.__class__.__name__,
            self.percent_coverage,
            self.ventilation.used,
            self.recirculation.used,
            self.dehumidification.used,
            self.panel.used,
        )

    def ToString(self):
        # type: () -> str
        return str(self)

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpCoolingParams
        new_obj = PhHeatPumpCoolingParams()
        new_obj.percent_coverage = self.percent_coverage
        new_obj.ventilation = self.ventilation.duplicate()
        new_obj.recirculation = self.recirculation.duplicate()
        new_obj.dehumidification = self.dehumidification.duplicate()
        new_obj.panel = self.panel.duplicate()
        return new_obj


# -----------------------------------------------------------------------------
# -- Heat Pumps Types


class PhHeatPumpAnnual(PhHeatPumpSystem):
    """Electric heat-pump with only Annual performance values."""

    def __init__(self):
        super(PhHeatPumpAnnual, self).__init__()
        self.annual_COP = 2.5
        self.total_system_perf_ratio = 0.4

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpAnnual, self).to_dict()
        d["annual_COP"] = self.annual_COP
        d["total_system_perf_ratio"] = self.total_system_perf_ratio
        d["cooling_params"] = self.cooling_params.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpAnnual
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.annual_COP = _input_dict["annual_COP"]
        new_obj.total_system_perf_ratio = _input_dict["total_system_perf_ratio"]
        new_obj.cooling_params = PhHeatPumpCoolingParams.from_dict(_input_dict["cooling_params"])
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpAnnual
        new_obj = PhHeatPumpAnnual()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.annual_COP = self.annual_COP
        new_obj.total_system_perf_ratio = self.total_system_perf_ratio
        new_obj.cooling_params = self.cooling_params.duplicate()
        return new_obj


class PhHeatPumpRatedMonthly(PhHeatPumpSystem):
    """Electric heat-pump with 2 separate monthly performance values."""

    def __init__(self):
        super(PhHeatPumpRatedMonthly, self).__init__()
        self.COP_1 = 2.5
        self.ambient_temp_1 = -8.333  # =17F
        self.COP_2 = 2.5
        self.ambient_temp_2 = 8.333  # =47F

    @property
    def monthly_COPS(self):
        return [self.COP_1, self.COP_2]

    @monthly_COPS.setter
    def monthly_COPS(self, _cops):
        # type: (Sequence[float]) -> None
        self.COP_1 = _cops[0]
        try:
            self.COP_2 = _cops[1]
        except IndexError:
            self.COP_2 = _cops[0]
        return

    @property
    def monthly_temps(self):
        return [self.ambient_temp_1, self.ambient_temp_2]

    @monthly_temps.setter
    def monthly_temps(self, _cops):
        # type: (Sequence[float]) -> None
        self.ambient_temp_1 = _cops[0]
        try:
            self.ambient_temp_2 = _cops[1]
        except IndexError:
            self.ambient_temp_2 = _cops[0]
        return

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = super(PhHeatPumpRatedMonthly, self).to_dict()
        d["COP_1"] = self.COP_1
        d["ambient_temp_1"] = self.ambient_temp_1
        d["COP_2"] = self.COP_2
        d["ambient_temp_2"] = self.ambient_temp_2
        d["cooling_params"] = self.cooling_params.to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpRatedMonthly
        new_obj = cls()
        new_obj.check_dict_type(_input_dict)
        new_obj.base_attrs_from_dict(_input_dict)
        new_obj.COP_1 = _input_dict["COP_1"]
        new_obj.ambient_temp_1 = _input_dict["ambient_temp_1"]
        new_obj.COP_2 = _input_dict["COP_2"]
        new_obj.ambient_temp_2 = _input_dict["ambient_temp_2"]
        new_obj.cooling_params = PhHeatPumpCoolingParams.from_dict(_input_dict["cooling_params"])
        return new_obj

    def __copy__(self):
        return self.duplicate()

    def duplicate(self):
        # type: () -> PhHeatPumpRatedMonthly
        new_obj = PhHeatPumpRatedMonthly()
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.COP_1 = self.COP_1
        new_obj.ambient_temp_1 = self.ambient_temp_1
        new_obj.COP_2 = self.COP_2
        new_obj.ambient_temp_2 = self.ambient_temp_2
        new_obj.cooling_params = self.cooling_params.duplicate()
        return new_obj


class PhHeatPumpCombined(PhHeatPumpSystem):
    def __init__(self):
        msg = "Sorry, Combined Heat Pumps are not yet supported in HBPH."
        raise NotImplementedError(msg)


# -----------------------------------------------------------------------------


class PhHeatPumpSystemBuilder(object):
    """Constructor class for HBPH-CoolingSystems"""

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHeatPumpSystem
        """Find the right heat-pump constructor class from the module based on the 'type' name."""

        valid_class_type_names = [nm for nm in dir(sys.modules[__name__]) if nm.startswith("PhHeatPump")]

        heat_pump_class_name = _input_dict["heat_pump_class_name"]
        if heat_pump_class_name not in valid_class_type_names:
            raise UnknownPhHeatPumpTypeError(valid_class_type_names, heat_pump_class_name)
        heat_pump_class = getattr(sys.modules[__name__], heat_pump_class_name)
        new_equipment = heat_pump_class.from_dict(_input_dict)
        return new_equipment

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-


"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


from honeybee_energy_ph.hvac import _base


class PhSupportiveDevice(_base._PhHVACBase):
    message = "PhSupportiveDevice is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Default data for Residential PH Electrical Appliances (PHI and Phius)."""

from honeybee_energy_ph.load.ph_equipment import (
    PhClothesDryerType,
    PhClothesWasherType,
    PhCookingType,
    PhDishwasherType,
)

ph_default_equip = {
    "PhDishwasher": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # Year
            "energy_demand": 0,  # kWh
            "energy_demand_per_use": 1.1,
            "combined_energy_factor": 0,
            "capacity_type": 1,  # Standard
            "capacity": 12,
            "water_connection": PhDishwasherType("2-COLD WATER CONNECTION"),
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # Year
            "energy_demand": 269,  # kWh
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "capacity_type": 1,  # Standard
            "capacity": 12,
            "water_connection": PhDishwasherType("2-COLD WATER CONNECTION"),
        },
    },
    "PhClothesWasher": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # Year
            "energy_demand": 0,
            "energy_demand_per_use": 1.1,
            "combined_energy_factor": 0,
            "capacity": 0.1274,
            "modified_energy_factor": 2.7,
            "water_connection": PhClothesWasherType("2-COLD WATER CONNECTION"),
            "utilization_factor": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # Year
            "energy_demand": 120,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "capacity": 0.1274,
            "modified_energy_factor": 2.7,
            "water_connection": PhClothesWasherType("2-COLD WATER CONNECTION"),
            "utilization_factor": 1.0,
        },
    },
    "PhClothesDryer": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # CEF - Combined Energy Factor
            "energy_demand": 0,
            "energy_demand_per_use": 3.5,
            "combined_energy_factor": 3.93,
            "dryer_type": PhClothesDryerType("4-CONDENSATION DRYER"),
            "gas_consumption": 0,
            "gas_efficiency_factor": 2.67,
            "field_utilization_factor_type": 1,  # Timer controls
            "field_utilization_factor": 1.18,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,  # CEF - Combined Energy Factor
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 3.93,
            "dryer_type": PhClothesDryerType("5-ELECTRIC EXHAUST AIR DRYER"),
            "gas_consumption": 0,
            "gas_efficiency_factor": 2.67,
            "field_utilization_factor_type": 1,  # Timer controls
            "field_utilization_factor": 1.18,
        },
    },
    "PhRefrigerator": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 0,
            "energy_demand_per_use": 0.78,
            "combined_energy_factor": 0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 1.0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
        },
    },
    "PhFreezer": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 0,
            "energy_demand_per_use": 0.88,
            "combined_energy_factor": 0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 0,
            "energy_demand_per_use": 0.88,
            "combined_energy_factor": 0,
        },
    },
    "PhFridgeFreezer": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 0,
            "energy_demand_per_use": 1.0,
            "combined_energy_factor": 0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 4,  # PH case Units
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Day
            "energy_demand": 1.22,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
        },
    },
    "PhCooktop": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0.0,  # kWh/use
            "energy_demand_per_use": 0.25,  # kWh/use
            "combined_energy_factor": 0,
            "cooktop_type": PhCookingType("1-ELECTRICITY"),
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 1,  # PH case occupants
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0.2,  # kWh/use
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "cooktop_type": PhCookingType("1-ELECTRICITY"),
        },
    },
    "PhPhiusMEL": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 3,  # Bedrooms
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 3,  # Bedrooms
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
        },
    },
    "PhPhiusLightingInterior": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 6,  # PH case floor area
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 6,  # PH case floor area
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
    "PhPhiusLightingExterior": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 6,  # PH case floor area
            "quantity": 1,
            "in_conditioned_space": False,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 6,  # PH case floor area
            "quantity": 1,
            "in_conditioned_space": False,
            "reference_energy_norm": 1,  # Use
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
    "PhPhiusLightingGarage": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 2,
            "quantity": 0,
            "in_conditioned_space": False,
            "reference_energy_norm": 2,
            "energy_demand": 100,
            "energy_demand_per_use": 100,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 2,
            "quantity": 0,
            "in_conditioned_space": False,
            "reference_energy_norm": 2,
            "energy_demand": 100,
            "energy_demand_per_use": 100,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
    "PhCustomAnnualElectric": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
    "PhCustomAnnualLighting": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
    "PhCustomAnnualMEL": {
        "PHI": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
        "PHIUS": {
            "comment": "default",
            "reference_quantity": 5,
            "quantity": 1,
            "in_conditioned_space": True,
            "reference_energy_norm": 2,
            "energy_demand": 0,
            "energy_demand_per_use": 0,
            "combined_energy_factor": 0,
            "frac_high_efficiency": 1.0,
        },
    },
}



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

try:
    from typing import TYPE_CHECKING, Any, Dict, Optional
except ImportError:
    TYPE_CHECKING = False
    pass  # Python 2.7

try:
    from ladybug_geometry import geometry3d
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from honeybee import properties
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    if TYPE_CHECKING:
        from honeybee_ph import space
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

"""
space.Space
  ├─ properties: SpaceProperties
      ├─ ph: SpacePhProperties
      ├─ energy: SpaceEnergyProperties
      ├─...
"""


class SpaceProperties(properties._Properties):
    """Honeybee-PH Space Properties.

    Space properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        space = Space()
        space.properties -> SpaceProperties
        space.properties.ph -> SpacePhProperties
        space.properties.energy -> SpaceEnergyProperties
    """

    def __init__(self, host):
        # type: (Optional[space.Space]) -> None
        self._host = host

    @property
    def host(self):
        # type: () -> Optional[space.Space]
        return self._host

    @property
    def host_name(self):
        # type: () -> str
        if self.host:
            return self.host.display_name
        else:
            return "None"

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        d = {}
        if abridged == False:
            d["type"] = "SpaceProperties"
        else:
            d["type"] = "SpacePropertiesAbridged"
        d = self._add_extension_attr_to_dict(d, abridged, include)
        return d

    @classmethod
    def from_dict(cls, _dict={}, _host=None):
        # type: (Dict, Any) -> SpaceProperties
        assert _dict["type"] == "SpaceProperties", "Expected SpaceProperties. Got {}.".format(_dict["type"])
        obj = cls(_host)
        return obj

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Space (eg. single-room HVAC systems) and does not add the
        prefix to attributes that are shared across several Spaces (eg. ConstructionSets).

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned at the level of this Room to default.

        This typically means erasing any ConstructionSets or ModifierSets assigned
        to this Space among other properties.
        """
        self._reset_extension_attr_to_default()

    def scale(self, factor, origin=None):
        # type: (float, Optional[geometry3d.Point3D]) -> None
        """Apply a scale transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be scaled alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        scaled if the host aperture is scaled.

        Arguments:
        ----------
            * factor (float): The scale factor
            * origin (Optional[geometry3d.Point3D]): default=None, A ladybug_geometry
                Point3D representing the origin from which to scale. If None,
                it will be scaled from the World origin (0, 0, 0).

        Returns:
        --------
            * None
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, "scale"):
                continue
            try:
                var.scale(factor, origin)
            except Exception as e:
                import traceback

                traceback.print_exc()
                raise Exception("Failed to scale {}: {}".format(var, e))

    def __repr__(self):
        """Properties representation."""
        return "{}(host={})".format(self.__class__.__name__, self.host_name)


class SpacePhProperties(object):
    def __init__(self, _host):
        # type: (Optional[SpaceProperties]) -> None
        self._host = _host
        self.id_num = 0

        # TODO: Temporary override until I figure out right right way
        self._v_sup = None
        self._v_eta = None
        self._v_tran = None

    @property
    def host(self):
        # type: () -> Optional[SpaceProperties]
        return self._host

    @property
    def host_name(self):
        # type: () -> str
        if self.host:
            return self.host.host_name
        else:
            return "None"

    @property
    def has_ventilation_flow_rates(self):
        # type: () -> bool
        """Return True if any of the Passive House ventilation flow rate overrides (sup, eta, trans) are set."""
        return any([self._v_sup, self._v_eta, self._v_tran])

    @property
    def honeybee_flow_rate(self):
        # type: () -> Optional[float]
        """Return the maximum of the Passive House ventilation flow rate overrides (sup, eta, trans), or None if none are set."""
        if not self.has_ventilation_flow_rates:
            return None
        return max([self._v_sup or 0, self._v_eta or 0, self._v_tran or 0])

    def duplicate(self, new_host=None):
        # type: (Any) -> SpacePhProperties
        _host = new_host or self._host
        new_properties_obj = SpacePhProperties(_host)

        new_properties_obj.id_num = self.id_num
        new_properties_obj._v_sup = self._v_sup
        new_properties_obj._v_eta = self._v_eta
        new_properties_obj._v_tran = self._v_tran

        return new_properties_obj

    def __str__(self):
        return "{}(host={}, v_eta={}, v_sup={}, v_tran={})".format(
            self.__class__.__name__,
            self.host_name,
            self._v_eta,
            self._v_sup,
            self._v_tran,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}

        if abridged:
            d["type"] = "SpacePhPropertiesAbridged"
        else:
            d["type"] = "SpacePhProperties"

        d["id_num"] = self.id_num
        d["_v_sup"] = self._v_sup
        d["_v_eta"] = self._v_eta
        d["_v_tran"] = self._v_tran

        return {"ph": d}

    @classmethod
    def from_dict(cls, data, host):
        # type: (dict, Any) -> SpacePhProperties
        assert "SpacePhProperties" in data["type"], "Expected SpacePhProperties. Got {}.".format(data["type"])
        new_prop = cls(host)

        new_prop.id_num = data["id_num"]
        new_prop._v_sup = data["_v_sup"]
        new_prop._v_eta = data["_v_eta"]
        new_prop._v_tran = data["_v_tran"]

        return new_prop


def get_ph_prop_from_space(_space):
    # type: (space.Space) -> SpacePhProperties
    """Get the space's PH-Properties."""
    return getattr(_space.properties, "ph")



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Package: Honeybee-HVAC-Equipment-PH

These classes are designed to be used within a Honeybee-editing environment, such as Rhino or 
Grasshopper. These classes will allow for Passive House specific HVAC data to be added to certain 
relevant Honeybee entities and logged for later use.
"""

# load all functions that extends honeybee core library
import honeybee_phhvac._extend_honeybee_ph_hvac
